@@ -26,6 +26,8 @@ import Control.Monad.Zip
 import Data.Data (Data)
 import Data.Either (lefts)
 import Data.Foldable as F hiding (concat, notElem)
+import Data.Function (on)
+import qualified Data.List as L
 import qualified Data.Map as M
 import Data.Map (Map)
 import Data.Maybe (isJust, mapMaybe)
@@ -260,6 +262,13 @@ dsExp (LamCasesE clauses) = do
   return $ DLamE args $ DCaseE (mkUnboxedTupleDExp (map DVarE args))
                                (map dClauseToUnboxedTupleMatch clauses')
 #endif
+#if __GLASGOW_HASKELL__ >= 907
+dsExp (TypedBracketE exp) = DTypedBracketE <$> dsExp exp
+dsExp (TypedSpliceE exp)  = DTypedSpliceE <$> dsExp exp
+#endif
+#if __GLASGOW_HASKELL__ >= 909
+dsExp (TypeE ty) = DTypeE <$> dsType ty
+#endif
 
 -- | Convert a 'DClause' to a 'DMatch' by bundling all of the clause's patterns
 -- into a match on a single unboxed tuple pattern. That is, convert this:
@@ -643,10 +652,14 @@ dsPat (SigP pat ty) = DSigP <$> dsPat pat <*> dsType ty
 dsPat (UnboxedSumP pat alt arity) =
   DConP (unboxedSumDataName alt arity) [] <$> ((:[]) <$> dsPat pat)
 #endif
+#if __GLASGOW_HASKELL__ >= 909
+dsPat (TypeP ty) = DTypeP <$> dsType ty
+dsPat (InvisP ty) = DInvisP <$> dsType ty
+#endif
 dsPat (ViewP _ _) =
   fail "View patterns are not supported in th-desugar. Use pattern guards instead."
 
--- | Convert a 'DPat' to a 'DExp'. Fails on 'DWildP'.
+-- | Convert a 'DPat' to a 'DExp'. Fails on 'DWildP' and 'DInvisP'.
 dPatToDExp :: DPat -> DExp
 dPatToDExp (DLitP lit) = DLitE lit
 dPatToDExp (DVarP name) = DVarE name
@@ -654,7 +667,9 @@ dPatToDExp (DConP name tys pats) = foldl DAppE (foldl DAppTypeE (DConE name) tys
 dPatToDExp (DTildeP pat) = dPatToDExp pat
 dPatToDExp (DBangP pat) = dPatToDExp pat
 dPatToDExp (DSigP pat ty) = DSigE (dPatToDExp pat) ty
+dPatToDExp (DTypeP ty) = DTypeE ty
 dPatToDExp DWildP = error "Internal error in th-desugar: wildcard in rhs of as-pattern"
+dPatToDExp (DInvisP {}) = error "Internal error in th-desugar: invisible type pattern in rhs of as-pattern"
 
 -- | Remove all wildcards from a pattern, replacing any wildcard with a fresh
 --   variable
@@ -665,6 +680,8 @@ removeWilds (DConP con_name tys pats) = DConP con_name tys <$> mapM removeWilds
 removeWilds (DTildeP pat) = DTildeP <$> removeWilds pat
 removeWilds (DBangP pat) = DBangP <$> removeWilds pat
 removeWilds (DSigP pat ty) = DSigP <$> removeWilds pat <*> pure ty
+removeWilds (DTypeP ty) = pure $ DTypeP ty
+removeWilds (DInvisP ty) = pure $ DInvisP ty
 removeWilds DWildP = DVarP <$> newUniqueName "wild"
 
 -- | Desugar @Info@
@@ -708,9 +725,9 @@ dsDec (DataD cxt n tvbs mk cons derivings) =
 dsDec (NewtypeD cxt n tvbs mk con derivings) =
   dsDataDec Newtype cxt n tvbs mk [con] derivings
 dsDec (TySynD n tvbs ty) =
-  (:[]) <$> (DTySynD n <$> mapM dsTvbUnit tvbs <*> dsType ty)
+  (:[]) <$> (DTySynD n <$> mapM dsTvbVis tvbs <*> dsType ty)
 dsDec (ClassD cxt n tvbs fds decs) =
-  (:[]) <$> (DClassD <$> dsCxt cxt <*> pure n <*> mapM dsTvbUnit tvbs
+  (:[]) <$> (DClassD <$> dsCxt cxt <*> pure n <*> mapM dsTvbVis tvbs
                      <*> pure fds <*> dsDecs decs)
 dsDec (InstanceD over cxt ty decs) =
   (:[]) <$> (DInstanceD over Nothing <$> dsCxt cxt <*> dsType ty <*> dsDecs decs)
@@ -721,7 +738,7 @@ dsDec d@(PragmaD {}) = dsTopLevelLetDec d
 dsDec (OpenTypeFamilyD tfHead) =
   (:[]) <$> (DOpenTypeFamilyD <$> dsTypeFamilyHead tfHead)
 dsDec (DataFamilyD n tvbs m_k) =
-  (:[]) <$> (DDataFamilyD n <$> mapM dsTvbUnit tvbs <*> mapM dsType m_k)
+  (:[]) <$> (DDataFamilyD n <$> mapM dsTvbVis tvbs <*> mapM dsType m_k)
 #if __GLASGOW_HASKELL__ >= 807
 dsDec (DataInstD cxt mtvbs lhs mk cons derivings) =
   case unfoldType lhs of
@@ -778,10 +795,10 @@ dsDec (TypeDataD n tys mk cons) =
 
 -- | Desugar a 'DataD', 'NewtypeD', or 'TypeDataD'.
 dsDataDec :: DsMonad q
-          => DataFlavor -> Cxt -> Name -> [TyVarBndrUnit]
+          => DataFlavor -> Cxt -> Name -> [TyVarBndrVis]
           -> Maybe Kind -> [Con] -> [DerivingClause] -> q [DDec]
 dsDataDec nd cxt n tvbs mk cons derivings = do
-  tvbs' <- mapM dsTvbUnit tvbs
+  tvbs' <- mapM dsTvbVis tvbs
   let h98_tvbs = case mk of
                    -- If there's an explicit return kind, we're dealing with a
                    -- GADT, so this argument goes unused in dsCon.
@@ -802,6 +819,7 @@ dsDataInstDec nd cxt n mtvbs tys mk cons derivings = do
   tys'   <- mapM dsTypeArg tys
   let lhs' = applyDType (DConT n) tys'
       h98_tvbs =
+        changeDTVFlags defaultBndrFlag $
         case (mk, mtvbs') of
           -- If there's an explicit return kind, we're dealing with a
           -- GADT, so this argument goes unused in dsCon.
@@ -827,7 +845,7 @@ dsFamilyResultSig (TyVarSig tvb) = DTyVarSig <$> dsTvbUnit tvb
 -- | Desugar a @TypeFamilyHead@
 dsTypeFamilyHead :: DsMonad q => TypeFamilyHead -> q DTypeFamilyHead
 dsTypeFamilyHead (TypeFamilyHead n tvbs result inj)
-  = DTypeFamilyHead n <$> mapM dsTvbUnit tvbs
+  = DTypeFamilyHead n <$> mapM dsTvbVis tvbs
                       <*> dsFamilyResultSig result
                       <*> pure inj
 
@@ -891,7 +909,13 @@ dsLetDec (ValD pat body where_decs) = do
 dsLetDec (SigD name ty) = do
   ty' <- dsType ty
   return ([DSigD name ty'], id)
-dsLetDec (InfixD fixity name) = return ([DInfixD fixity name], id)
+#if __GLASGOW_HASKELL__ >= 909
+dsLetDec (InfixD fixity ns_spec name) =
+  return ([DInfixD fixity ns_spec name], id)
+#else
+dsLetDec (InfixD fixity name) =
+  return ([DInfixD fixity NoNamespaceSpecifier name], id)
+#endif
 dsLetDec (PragmaD prag) = do
   prag' <- dsPragma prag
   return ([DPragmaD prag'], id)
@@ -945,10 +969,10 @@ dsTopLevelLetDec = fmap (map DLetDec . fst) . dsLetDec
 -- we require passing these as arguments. (If we desugar an actual GADT
 -- constructor, these arguments are ignored.)
 dsCon :: DsMonad q
-      => [DTyVarBndrUnit] -- ^ The universally quantified type variables
-                          --   (used if desugaring a non-GADT constructor).
-      -> DType            -- ^ The original data declaration's type
-                          --   (used if desugaring a non-GADT constructor).
+      => [DTyVarBndrVis] -- ^ The universally quantified type variables
+                         --   (used if desugaring a non-GADT constructor).
+      -> DType           -- ^ The original data declaration's type
+                         --   (used if desugaring a non-GADT constructor).
       -> Con -> q [DCon]
 dsCon univ_dtvbs data_type con = do
   dcons' <- dsCon' con
@@ -959,7 +983,7 @@ dsCon univ_dtvbs data_type con = do
             expl_dtvbs = changeDTVFlags SpecifiedSpec univ_dtvbs ++
                          ex_dtvbs
             impl_dtvbs = changeDTVFlags SpecifiedSpec $
-                         toposortTyVarsOf $ mapMaybe extractTvbKind expl_dtvbs in
+                         toposortKindVarsOfTvbs expl_dtvbs in
         DCon (impl_dtvbs ++ expl_dtvbs) dcxt n fields data_type
       Just gadt_type ->
         let univ_ex_dtvbs = dtvbs in
@@ -1052,6 +1076,9 @@ dsPragma (CompleteP cls mty)             = return $ DCompleteP cls mty
 #if __GLASGOW_HASKELL__ >= 903
 dsPragma (OpaqueP n)                     = return $ DOpaqueP n
 #endif
+#if __GLASGOW_HASKELL__ >= 909
+dsPragma (SCCP nm mstr)                  = return $ DSCCP nm mstr
+#endif
 
 -- | Desugar a @RuleBndr@.
 dsRuleBndr :: DsMonad q => RuleBndr -> q DRuleBndr
@@ -1289,6 +1316,14 @@ dsTvbUnit = dsTvb
 dsTvbUnit = dsTvb ()
 #endif
 
+-- | Desugar a 'TyVarBndrVis'.
+dsTvbVis :: DsMonad q => TyVarBndrVis -> q DTyVarBndrVis
+#if __GLASGOW_HASKELL__ >= 900
+dsTvbVis = dsTvb
+#else
+dsTvbVis = dsTvb BndrReq
+#endif
+
 -- | Desugar a @Cxt@
 dsCxt :: DsMonad q => Cxt -> q DCxt
 dsCxt = concatMapM dsPred
@@ -1505,6 +1540,8 @@ isUniversalPattern (DTildeP {})  = return True
 isUniversalPattern (DBangP pat)  = isUniversalPattern pat
 isUniversalPattern (DSigP pat _) = isUniversalPattern pat
 isUniversalPattern DWildP        = return True
+isUniversalPattern (DTypeP _)    = return True
+isUniversalPattern (DInvisP _)   = return True
 
 -- | Apply one 'DExp' to a list of arguments
 applyDExp :: DExp -> [DExp] -> DExp
@@ -1546,6 +1583,38 @@ dTyVarBndrToDType :: DTyVarBndr flag -> DType
 dTyVarBndrToDType (DPlainTV a _)    = DVarT a
 dTyVarBndrToDType (DKindedTV a _ k) = DVarT a `DSigT` k
 
+-- | Convert a 'DTyVarBndrVis' to a 'DTypeArg'. That is, convert a binder with a
+-- 'BndrReq' visibility to a 'DTANormal' and a binder with 'BndrInvis'
+-- visibility to a 'DTyArg'.
+--
+-- If given a 'DKindedTV', the resulting 'DTypeArg' will omit the kind
+-- signature. Use 'dTyVarBndrVisToDTypeArgWithSig' if you want to preserve the
+-- kind signature.
+dTyVarBndrVisToDTypeArg :: DTyVarBndrVis -> DTypeArg
+dTyVarBndrVisToDTypeArg bndr =
+  case dtvbFlag bndr of
+    BndrReq   -> DTANormal bndr_ty
+    BndrInvis -> DTyArg bndr_ty
+  where
+    bndr_ty = case bndr of
+                DPlainTV a _    -> DVarT a
+                DKindedTV a _ _ -> DVarT a
+
+-- | Convert a 'DTyVarBndrVis' to a 'DTypeArg'. That is, convert a binder with a
+-- 'BndrReq' visibility to a 'DTANormal' and a binder with 'BndrInvis'
+-- visibility to a 'DTyArg'.
+--
+-- If given a 'DKindedTV', the resulting 'DTypeArg' will preserve the kind
+-- signature. Use 'dTyVarBndrVisToDTypeArg' if you want to omit the kind
+-- signature.
+dTyVarBndrVisToDTypeArgWithSig :: DTyVarBndrVis -> DTypeArg
+dTyVarBndrVisToDTypeArgWithSig bndr =
+  case dtvbFlag bndr of
+    BndrReq   -> DTANormal bndr_ty
+    BndrInvis -> DTyArg bndr_ty
+  where
+    bndr_ty = dTyVarBndrToDType bndr
+
 -- | Extract the underlying 'DType' or 'DKind' from a 'DTypeArg'. This forgets
 -- information about whether a type is a normal argument or not, so use with
 -- caution.
@@ -1555,9 +1624,9 @@ probablyWrongUnDTypeArg (DTyArg k)    = k
 
 -- Take a data type name (which does not belong to a data family) and
 -- apply it to its type variable binders to form a DType.
-nonFamilyDataReturnType :: Name -> [DTyVarBndrUnit] -> DType
+nonFamilyDataReturnType :: Name -> [DTyVarBndrVis] -> DType
 nonFamilyDataReturnType con_name =
-  applyDType (DConT con_name) . map (DTANormal . dTyVarBndrToDType)
+  applyDType (DConT con_name) . map dTyVarBndrVisToDTypeArg
 
 -- Take a data family name and apply it to its argument types to form a
 -- data family instance DType.
@@ -1679,10 +1748,33 @@ toposortTyVarsOf tys =
   in map ascribeWithKind $
      scopedSort freeVars
 
+-- | Take a telescope of 'DTyVarBndr's, find the free variables in their kinds,
+-- and sort them in reverse topological order to ensure that they are well
+-- scoped. Because the argument list is assumed to be telescoping, kind
+-- variables that are bound earlier in the list are not returned. For example,
+-- this:
+--
+-- @
+-- 'toposortKindVarsOfTvbs' [a :: k, b :: Proxy a]
+-- @
+--
+-- Will return @[k]@, not @[k, a]@, since @a@ is bound earlier by @a :: k@.
+toposortKindVarsOfTvbs :: [DTyVarBndr flag] -> [DTyVarBndrUnit]
+toposortKindVarsOfTvbs tvbs =
+  foldr (\tvb kvs ->
+          foldMap (\t -> toposortTyVarsOf [t]) (extractTvbKind tvb) `L.union`
+          L.deleteBy ((==) `on` dtvbName) tvb kvs)
+        []
+        (changeDTVFlags () tvbs)
+
 dtvbName :: DTyVarBndr flag -> Name
 dtvbName (DPlainTV n _)    = n
 dtvbName (DKindedTV n _ _) = n
 
+dtvbFlag :: DTyVarBndr flag -> flag
+dtvbFlag (DPlainTV _ flag)    = flag
+dtvbFlag (DKindedTV _ flag _) = flag
+
 -- @mk_qual_do_name mb_mod orig_name@ will simply return @orig_name@ if
 -- @mb_mod@ is Nothing. If @mb_mod@ is @Just mod_@, then a new 'Name' will be
 -- returned that uses @mod_@ as the new module prefix. This is useful for
