@@ -24,7 +24,7 @@ module Language.Haskell.TH.Desugar.Subst (
   IgnoreKinds(..), matchTy
   ) where
 
-import Data.List
+import qualified Data.List as L
 import qualified Data.Map as M
 import qualified Data.Set as S
 
@@ -32,10 +32,6 @@ import Language.Haskell.TH.Desugar.AST
 import Language.Haskell.TH.Syntax
 import Language.Haskell.TH.Desugar.Util
 
-#if __GLASGOW_HASKELL__ < 710
-import Control.Applicative
-#endif
-
 -- | A substitution is just a map from names to types
 type DSubst = M.Map Name DType
 
@@ -120,12 +116,17 @@ data IgnoreKinds = YesIgnore | NoIgnore
 matchTy :: IgnoreKinds -> DType -> DType -> Maybe DSubst
 matchTy _   (DVarT var_name) arg = Just $ M.singleton var_name arg
   -- if a pattern has a kind signature, it's really easy to get
-  -- this wrong.
+  -- the following two cases wrong.
 matchTy ign (DSigT ty _ki) arg = case ign of
   YesIgnore -> matchTy ign ty arg
   NoIgnore  -> Nothing
-  -- but we can safely ignore kind signatures on the target
-matchTy ign pat (DSigT ty _ki) = matchTy ign pat ty
+matchTy ign (DAppKindT ty _ki) arg = case ign of
+  YesIgnore -> matchTy ign ty arg
+  NoIgnore  -> Nothing
+  -- but we can safely ignore kind signatures on the target,
+  -- as in the following two cases.
+matchTy ign pat (DSigT     ty _ki) = matchTy ign pat ty
+matchTy ign pat (DAppKindT ty _ki) = matchTy ign pat ty
 matchTy _   (DForallT {}) _ =
   error "Cannot match a forall in a pattern"
 matchTy _   _ (DForallT {}) =
@@ -140,7 +141,7 @@ matchTy _   (DLitT pat_lit) (DLitT arg_lit)
 matchTy _ _ _ = Nothing
 
 unionMaybeSubsts :: [Maybe DSubst] -> Maybe DSubst
-unionMaybeSubsts = foldl' union_subst1 (Just M.empty)
+unionMaybeSubsts = L.foldl' union_subst1 (Just M.empty)
   where
     union_subst1 :: Maybe DSubst -> Maybe DSubst -> Maybe DSubst
     union_subst1 ma mb = do
