@@ -8,7 +8,7 @@ Utility functions for th-desugar package.
 
 {-# LANGUAGE CPP, DeriveDataTypeable, DeriveGeneric, DeriveLift, RankNTypes,
              ScopedTypeVariables, TupleSections, AllowAmbiguousTypes,
-             TemplateHaskellQuotes, TypeApplications #-}
+             TemplateHaskellQuotes, TypeApplications, MagicHash #-}
 
 module Language.Haskell.TH.Desugar.Util (
   newUniqueName,
@@ -19,16 +19,16 @@ module Language.Haskell.TH.Desugar.Util (
   stripPlainTV_maybe,
   thirdOf3, splitAtList, extractBoundNamesDec,
   extractBoundNamesPat,
-  tvbToType, tvbToTypeWithSig, tvbToTANormalWithSig,
+  tvbToType, tvbToTypeWithSig,
   nameMatches, thdOf3, liftFst, liftSnd, firstMatch, firstMatchM,
-  unboxedSumDegree_maybe, unboxedSumNameDegree_maybe,
-  tupleDegree_maybe, tupleNameDegree_maybe, unboxedTupleDegree_maybe,
-  unboxedTupleNameDegree_maybe, splitTuple_maybe,
+  tupleNameDegree_maybe,
+  unboxedSumNameDegree_maybe, unboxedTupleNameDegree_maybe, splitTuple_maybe,
   topEverywhereM, isInfixDataCon,
   isTypeKindName, typeKindName,
   unSigType, unfoldType, ForallTelescope(..), FunArgs(..), VisFunArg(..),
   filterVisFunArgs, ravelType, unravelType,
   TypeArg(..), applyType, filterTANormals, probablyWrongUnTypeArg,
+  tyVarBndrVisToTypeArg, tyVarBndrVisToTypeArgWithSig,
   bindIP,
   DataFlavor(..)
   ) where
@@ -51,6 +51,21 @@ import GHC.Classes ( IP )
 import GHC.Generics ( Generic )
 import Unsafe.Coerce ( unsafeCoerce )
 
+#if __GLASGOW_HASKELL__ >= 906
+import GHC.Tuple ( Solo(MkSolo) )
+#elif __GLASGOW_HASKELL__ >= 900
+import GHC.Tuple ( Solo(Solo) )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 908
+import GHC.Tuple ( Tuple0, Unit )
+import Text.Read ( readMaybe )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 910
+import GHC.Types ( Solo#, Sum2#, Tuple0#, Unit# )
+#endif
+
 ----------------------------------------
 -- TH manipulations
 ----------------------------------------
@@ -117,11 +132,6 @@ tvbToType = VarT . tvName
 tvbToTypeWithSig :: TyVarBndr_ flag -> Type
 tvbToTypeWithSig = elimTV VarT (\n k -> SigT (VarT n) k)
 
--- | Convert a 'TyVarBndr' into a 'TypeArg' (specifically, a 'TANormal'),
--- preserving the kind signature (if it has one).
-tvbToTANormalWithSig :: TyVarBndr_ flag -> TypeArg
-tvbToTANormalWithSig = TANormal . tvbToTypeWithSig
-
 -- | Do two names name the same thing?
 nameMatches :: Name -> Name -> Bool
 nameMatches n1@(Name occ1 flav1) n2@(Name occ2 flav2)
@@ -139,27 +149,103 @@ nameMatches n1@(Name occ1 flav1) n2@(Name occ2 flav2)
   | otherwise
   = n1 == n2
 
--- | Extract the degree of a tuple
-tupleDegree_maybe :: String -> Maybe Int
-tupleDegree_maybe s = do
-  '(' : s1 <- return s
-  (commas, ")") <- return $ span (== ',') s1
-  let degree
-        | "" <- commas = 0
-        | otherwise    = length commas + 1
-  return degree
-
--- | Extract the degree of a tuple name
+-- | Extract the degree of a tuple 'Name'.
+--
+-- In addition to recognizing tuple syntax (e.g., @''(,,)@), this also
+-- recognizes the following:
+--
+-- * @''Unit@ (for 0-tuples)
+--
+-- * @''Solo@/@'MkSolo@ (for 1-tuples)
+--
+-- * @''Tuple<N>@ (for <N>-tuples)
+--
+-- In recent versions of GHC, @''()@ is a synonym for @''Unit@, @''(,)@ is a
+-- synonym for @''Tuple2@, and so on. As a result, we must check for @''Unit@
+-- and @''Tuple<N>@ in 'tupleDegree_maybe' to be thorough. (There is no special
+-- tuple syntax for @''Solo@/@'MkSolo@, but we check them here as well for the
+-- sake of completeness.)
 tupleNameDegree_maybe :: Name -> Maybe Int
-tupleNameDegree_maybe = tupleDegree_maybe . nameBase
+tupleNameDegree_maybe name
+  -- First, check for Solo/MkSolo...
+#if __GLASGOW_HASKELL__ >= 900
+  | name == ''Solo = Just 1
+#if __GLASGOW_HASKELL__ >= 906
+  | name == 'MkSolo = Just 1
+#else
+  | name == 'Solo = Just 1
+#endif
+#endif
+#if __GLASGOW_HASKELL__ >= 908
+  -- ...then, check for Unit...
+  | name == ''Unit = Just 0
+  -- ...then, check for Tuple<N>. It is theoretically possible for the supplied
+  -- Name to be a user-defined data type named Tuple<N>, rather than the actual
+  -- tuple types defined in GHC.Tuple. As such, we check that the package and
+  -- module for the supplied Name does in fact come from ghc-prim:GHC.Tuple as
+  -- a sanity check.
+  | -- We use Tuple0 here simply because it is a convenient identifier from
+    -- GHC.Tuple. We could just as well use any other identifier from GHC.Tuple,
+    -- however.
+    namePackage name == namePackage ''Tuple0
+  , nameModule name == nameModule ''Tuple0
+  , 'T':'u':'p':'l':'e':n <- nameBase name
+    -- This relies on the Read Int instance. This is more permissive than what
+    -- we need, since there are valid Int strings (e.g., "-1") that do not have
+    -- corresponding Tuple<N> names (e.g., "Tuple-1" is not a data type in
+    -- GHC.Tuple). As such, we depend on the assumption that the input string
+    -- does in fact come from GHC.Tuple, which we check above.
+  = readMaybe n
+#endif
+  -- ...otherwise, fall back on tuple syntax.
+  | otherwise
+  = tuple_syntax_degree_maybe (nameBase name)
+  where
+    -- Extract the degree of a string using tuple syntax (e.g., @''(,,)@).
+    tuple_syntax_degree_maybe :: String -> Maybe Int
+    tuple_syntax_degree_maybe s = do
+      '(' : s1 <- return s
+      (commas, ")") <- return $ span (== ',') s1
+      let degree
+            | "" <- commas = 0
+            | otherwise    = length commas + 1
+      return degree
 
 -- | Extract the degree of an unboxed sum
 unboxedSumDegree_maybe :: String -> Maybe Int
 unboxedSumDegree_maybe = unboxedSumTupleDegree_maybe '|'
 
--- | Extract the degree of an unboxed sum name
+-- | Extract the degree of an unboxed sum 'Name'.
+--
+-- In addition to recognizing unboxed sum syntax (e.g., @''(#||#)@), this also
+-- recognizes @''Sum<N>#@ (for unboxed <N>-ary sum type constructors). In recent
+-- versions of GHC, @''Sum2#@ is a synonym for @''(#|#)@, @''Sum3#@ is a synonym
+-- for @''(#||#)@, and so on. As a result, we must check for @''Sum<N>#@ in
+-- 'unboxedSumNameDegree_maybe' to be thorough.
 unboxedSumNameDegree_maybe :: Name -> Maybe Int
-unboxedSumNameDegree_maybe = unboxedSumDegree_maybe . nameBase
+unboxedSumNameDegree_maybe name
+#if __GLASGOW_HASKELL__ >= 910
+  -- Check for Sum<N>#. It is theoretically possible for the supplied
+  -- Name to be a user-defined data type named Sum<N>#, rather than the actual
+  -- unboxed sum types defined in GHC.Types. As such, we check that the package
+  -- and module for the supplied Name does in fact come from ghc-prim:GHC.Types
+  -- as a sanity check.
+  | -- We use Sum2# here simply because it is a convenient identifier from
+    -- GHC.Types. We could just as well use any other identifier from GHC.Types,
+    -- however.
+    namePackage name == namePackage ''Sum2#
+  , nameModule name == nameModule ''Sum2#
+  , 'S':'u':'m':n:"#" <- nameBase name
+    -- This relies on the Read Int instance. This is more permissive than what
+    -- we need, since there are valid Int strings (e.g., "-1") that do not have
+    -- corresponding Sum<N># names (e.g., "Sum-1#" is not a data type in
+    -- GHC.Types). As such, we depend on the assumption that the input string
+    -- does in fact come from GHC.Types, which we check above.
+  = readMaybe [n]
+#endif
+  -- ...otherwise, fall back on unboxed sum syntax.
+  | otherwise
+  = unboxedSumDegree_maybe (nameBase name)
 
 -- | Extract the degree of an unboxed tuple
 unboxedTupleDegree_maybe :: String -> Maybe Int
@@ -175,9 +261,50 @@ unboxedSumTupleDegree_maybe sep s = do
         | otherwise  = length seps + 1
   return degree
 
--- | Extract the degree of an unboxed tuple name
+-- | Extract the degree of an unboxed tuple 'Name'.
+--
+-- In addition to recognizing unboxed tuple syntax (e.g., @''(#,,#)@), this also
+-- recognizes the following:
+--
+-- * @''Unit#@ (for unboxed 0-tuples)
+--
+-- * @''Solo#@/@'Solo#@ (for unboxed 1-tuples)
+--
+-- * @''Tuple<N>#@ (for unboxed <N>-tuples)
+--
+-- In recent versions of GHC, @''(##)@ is a synonym for @''Unit#@, @''(#,#)@ is
+-- a synonym for @''Tuple2#@, and so on. As a result, we must check for
+-- @''Unit#@, and @''Tuple<N>@ in 'unboxedTupleNameDegree_maybe' to be thorough.
+-- (There is no special unboxed tuple type constructor for @''Solo#@/@'Solo#@,
+-- but we check them here as well for the sake of completeness.)
 unboxedTupleNameDegree_maybe :: Name -> Maybe Int
-unboxedTupleNameDegree_maybe = unboxedTupleDegree_maybe . nameBase
+unboxedTupleNameDegree_maybe name
+#if __GLASGOW_HASKELL__ >= 910
+  -- First, check for Solo#...
+  | name == ''Solo# = Just 1
+  -- ...then, check for Unit#...
+  | name == ''Unit# = Just 0
+  -- ...then, check for Tuple<N>#. It is theoretically possible for the supplied
+  -- Name to be a user-defined data type named Tuple<N>#, rather than the actual
+  -- unboxed tuple types defined in GHC.Types. As such, we check that the
+  -- package and module for the supplied Name does in fact come from
+  -- ghc-prim:GHC.Types as a sanity check.
+  | -- We use Tuple0# here simply because it is a convenient identifier from
+    -- GHC.Types. We could just as well use any other identifier from GHC.Types,
+    -- however.
+    namePackage name == namePackage ''Tuple0#
+  , nameModule name == nameModule ''Tuple0#
+  , 'T':'u':'p':'l':'e':n:"#" <- nameBase name
+    -- This relies on the Read Int instance. This is more permissive than what
+    -- we need, since there are valid Int strings (e.g., "-1") that do not have
+    -- corresponding Tuple<N># names (e.g., "Tuple-1#" is not a data type in
+    -- GHC.Types). As such, we depend on the assumption that the input string
+    -- does in fact come from GHC.Types, which we check above.
+  = readMaybe [n]
+#endif
+  -- ...otherwise, fall back on unboxed tuple syntax.
+  | otherwise
+  = unboxedTupleDegree_maybe (nameBase name)
 
 -- | If the argument is a tuple type, return the components
 splitTuple_maybe :: Type -> Maybe [Type]
@@ -310,18 +437,40 @@ unSigPred = unSigType
 -- @
 -- ('ConT' ''Proxy, ['TyArg' ('ConT' ''Type), 'TANormal' ('ConT' ''Char)])
 -- @
+--
+-- This process forgets about infix application, so both of these types:
+--
+-- @
+-- Int :++: Int
+-- (:++:) Int Int
+-- @
+--
+-- will be unfolded to this:
+--
+-- @
+-- ('ConT' ''(:+:), ['TANormal' ('ConT' ''Int), 'TANormal' ('ConT' ''Int)])
+-- @
+--
+-- This function should only be used after all 'UInfixT' and 'PromotedUInfixT'
+-- types have been resolved (e.g., via @th-abstraction@'s
+-- @<https://hackage.haskell.org/package/th-abstraction-0.5.0.0/docs/Language-Haskell-TH-Datatype.html#v:resolveInfixT resolveInfixT>@
+-- function).
 unfoldType :: Type -> (Type, [TypeArg])
 unfoldType = go []
   where
     go :: [TypeArg] -> Type -> (Type, [TypeArg])
-    go acc (ForallT _ _ ty) = go acc ty
-    go acc (AppT ty1 ty2)   = go (TANormal ty2:acc) ty1
-    go acc (SigT ty _)      = go acc ty
-    go acc (ParensT ty)     = go acc ty
+    go acc (ForallT _ _ ty)           = go acc ty
+    go acc (AppT ty1 ty2)             = go (TANormal ty2:acc) ty1
+    go acc (SigT ty _)                = go acc ty
+    go acc (ParensT ty)               = go acc ty
+    go acc (InfixT ty1 n ty2)         = go (TANormal ty1:TANormal ty2:acc) (ConT n)
 #if __GLASGOW_HASKELL__ >= 807
-    go acc (AppKindT ty ki) = go (TyArg ki:acc) ty
+    go acc (AppKindT ty ki)           = go (TyArg ki:acc) ty
 #endif
-    go acc ty               = (ty, acc)
+#if __GLASGOW_HASKELL__ >= 904
+    go acc (PromotedInfixT ty1 n ty2) = go (TANormal ty1:TANormal ty2:acc) (PromotedT n)
+#endif
+    go acc ty                         = (ty, acc)
 
 -- | An argument to a type, either a normal type ('TANormal') or a visible
 -- kind application ('TyArg').
@@ -356,6 +505,36 @@ filterTANormals = mapMaybe getTANormal
     getTANormal (TANormal t) = Just t
     getTANormal (TyArg {})   = Nothing
 
+-- | Convert a 'TyVarBndrVis' to a 'TypeArg'. That is, convert a binder with a
+-- 'BndrReq' visibility to a 'TANormal' and a binder with 'BndrInvis'
+-- visibility to a 'TyArg'.
+--
+-- If given a 'KindedTV', the resulting 'TypeArg' will omit the kind signature.
+-- Use 'tyVarBndrVisToTypeArgWithSig' if you want to preserve the kind
+-- signature.
+tyVarBndrVisToTypeArg :: TyVarBndrVis -> TypeArg
+tyVarBndrVisToTypeArg bndr =
+  case tvFlag bndr of
+    BndrReq   -> TANormal bndr_ty
+    BndrInvis -> TyArg bndr_ty
+  where
+    bndr_ty = tvbToType bndr
+
+-- | Convert a 'TyVarBndrVis' to a 'TypeArg'. That is, convert a binder with a
+-- 'BndrReq' visibility to a 'TANormal' and a binder with 'BndrInvis'
+-- visibility to a 'TyArg'.
+--
+-- If given a 'KindedTV', the resulting 'TypeArg' will preserve the kind
+-- signature. Use 'tyVarBndrVisToTypeArg' if you want to omit the kind
+-- signature.
+tyVarBndrVisToTypeArgWithSig :: TyVarBndrVis -> TypeArg
+tyVarBndrVisToTypeArgWithSig bndr =
+  case tvFlag bndr of
+    BndrReq   -> TANormal bndr_ty
+    BndrInvis -> TyArg bndr_ty
+  where
+    bndr_ty = tvbToTypeWithSig bndr
+
 -- | Extract the underlying 'Type' or 'Kind' from a 'TypeArg'. This forgets
 -- information about whether a type is a normal argument or not, so use with
 -- caution.
@@ -375,7 +554,10 @@ nameOccursIn n = everything (||) $ mkQ False (== n)
 allNamesIn :: Data a => a -> [Name]
 allNamesIn = everything (++) $ mkQ [] (:[])
 
--- | Extract the names bound in a @Stmt@
+-- | Extract the names bound in a @Stmt@.
+--
+-- This does /not/ extract any type variables bound by pattern signatures,
+-- constructor patterns, or type patterns.
 extractBoundNamesStmt :: Stmt -> OSet Name
 extractBoundNamesStmt (BindS pat _) = extractBoundNamesPat pat
 extractBoundNamesStmt (LetS decs)   = foldMap extractBoundNamesDec decs
@@ -386,12 +568,18 @@ extractBoundNamesStmt (RecS stmtss) = foldMap extractBoundNamesStmt stmtss
 #endif
 
 -- | Extract the names bound in a @Dec@ that could appear in a @let@ expression.
+--
+-- This does /not/ extract any type variables bound by pattern signatures,
+-- constructor patterns, or type patterns.
 extractBoundNamesDec :: Dec -> OSet Name
 extractBoundNamesDec (FunD name _)  = OS.singleton name
 extractBoundNamesDec (ValD pat _ _) = extractBoundNamesPat pat
 extractBoundNamesDec _              = OS.empty
 
--- | Extract the names bound in a @Pat@
+-- | Extract the names bound in a @Pat@.
+--
+-- This does /not/ extract any type variables bound by pattern signatures,
+-- constructor patterns, or type patterns.
 extractBoundNamesPat :: Pat -> OSet Name
 extractBoundNamesPat (LitP _)              = OS.empty
 extractBoundNamesPat (VarP name)           = OS.singleton name
@@ -420,6 +608,10 @@ extractBoundNamesPat (ViewP _ pat)         = extractBoundNamesPat pat
 #if __GLASGOW_HASKELL__ >= 801
 extractBoundNamesPat (UnboxedSumP pat _ _) = extractBoundNamesPat pat
 #endif
+#if __GLASGOW_HASKELL__ >= 909
+extractBoundNamesPat (TypeP _)             = OS.empty
+extractBoundNamesPat (InvisP _)            = OS.empty
+#endif
 
 ----------------------------------------
 -- General utility
