@@ -9,6 +9,12 @@ rae@cs.brynmawr.edu
              FlexibleInstances, DataKinds, CPP, RankNTypes,
              StandaloneDeriving, DefaultSignatures,
              ConstraintKinds, RoleAnnotations, DeriveAnyClass #-}
+#if __GLASGOW_HASKELL__ >= 810
+{-# LANGUAGE StandaloneKindSignatures #-}
+#endif
+#if __GLASGOW_HASKELL__ >= 907
+{-# LANGUAGE TypeAbstractions #-}
+#endif
 
 {-# OPTIONS_GHC -Wno-orphans -Wno-name-shadowing
                 -Wno-redundant-constraints #-}
@@ -44,6 +50,15 @@ $(S.dectest17)
 $(S.dectest18)
 #endif
 
+#if __GLASGOW_HASKELL__ >= 907
+$(S.dectest19)
+#endif
+
+#if __GLASGOW_HASKELL__ >= 909
+$(S.dectest20)
+$(S.dectest21)
+#endif
+
 $(fmap unqualify S.instance_test)
 
 $(fmap unqualify S.imp_inst_test1)
