@@ -10,7 +10,7 @@ rae@cs.brynmawr.edu
              FlexibleInstances, ExistentialQuantification,
              ScopedTypeVariables, GADTs, ViewPatterns, TupleSections,
              TypeOperators, PartialTypeSignatures, PatternSynonyms,
-             TypeApplications #-}
+             TypeApplications, MagicHash #-}
 {-# OPTIONS -Wno-incomplete-patterns -Wno-overlapping-patterns
             -Wno-unused-matches -Wno-type-defaults
             -Wno-missing-signatures -Wno-unused-do-bind
@@ -22,6 +22,10 @@ rae@cs.brynmawr.edu
 {-# LANGUAGE QuantifiedConstraints #-}
 #endif
 
+#if __GLASGOW_HASKELL__ < 806
+{-# LANGUAGE TypeInType #-}
+#endif
+
 #if __GLASGOW_HASKELL__ >= 809
 {-# LANGUAGE StandaloneKindSignatures #-}
 #endif
@@ -30,6 +34,14 @@ rae@cs.brynmawr.edu
 {-# LANGUAGE TypeData #-}
 #endif
 
+#if __GLASGOW_HASKELL__ >= 907
+{-# LANGUAGE TypeAbstractions #-}
+#endif
+
+#if __GLASGOW_HASKELL__ >= 909
+{-# LANGUAGE RequiredTypeArguments #-}
+#endif
+
 module Main where
 
 import Prelude hiding ( exp )
@@ -45,11 +57,13 @@ import Dec ( RecordSel )
 import ReifyTypeCUSKs
 import ReifyTypeSigs
 import T159Decs ( t159A, t159B )
+import T183 ( t183 )
+import qualified Language.Haskell.TH.Datatype.TyVarBndr as THAbs
 import Language.Haskell.TH.Desugar
 import qualified Language.Haskell.TH.Desugar.OSet as OS
 import Language.Haskell.TH.Desugar.Expand  ( expandUnsoundly )
 import Language.Haskell.TH
-import qualified Language.Haskell.TH.Syntax as Syn ( NameSpace(..), lift )
+import qualified Language.Haskell.TH.Syntax as Syn
 
 import Control.Exception ( ErrorCall )
 import Control.Monad
@@ -61,6 +75,22 @@ import Data.Proxy
 import Prelude as P
 #endif
 
+#if __GLASGOW_HASKELL__ >= 906
+import GHC.Tuple ( Solo(MkSolo) )
+#elif __GLASGOW_HASKELL__ >= 900
+import GHC.Tuple ( Solo(Solo) )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 908
+import qualified FakeTuples
+import GHC.Tuple ( Tuple0, Tuple1, Tuple2, Tuple3, Unit )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 910
+import qualified FakeSums
+import GHC.Types (Solo#, Sum2#, Sum3#, Sum4#, Tuple0#, Tuple1#, Tuple2#, Tuple3#, Unit#)
+#endif
+
 -- |
 -- Convert a HUnit test suite to a spec.  This can be used to run existing
 -- HUnit tests with Hspec.
@@ -154,6 +184,15 @@ tests = test [ "sections" ~: $test1_sections  @=? $(dsSplice test1_sections)
 #if __GLASGOW_HASKELL__ >= 903
              , "opaque_pragma" ~: $test55_opaque_pragma @=? $(dsSplice test55_opaque_pragma)
              , "lambda_cases" ~: $test56_lambda_cases @=? $(dsSplice test56_lambda_cases)
+#endif
+#if __GLASGOW_HASKELL__ >= 907
+             , "typed_th_bracket" ~: $$($test57_typed_th_bracket) @=? $$($(dsSplice test57_typed_th_bracket))
+             , "typed_th_splice" ~: $test58_typed_th_splice @=? $(dsSplice test58_typed_th_splice)
+#endif
+#if __GLASGOW_HASKELL__ >= 909
+             , "embedded_types_keyword" ~: $test59_embedded_types_keyword @=? $(dsSplice test59_embedded_types_keyword)
+             , "embedded_types_no_keyword" ~: $test60_embedded_types_no_keyword @=? $(dsSplice test60_embedded_types_no_keyword)
+             , "invis_type_pat" ~: $test61_invis_type_pat @=? $(dsSplice test61_invis_type_pat)
 #endif
              ]
 
@@ -276,7 +315,7 @@ test_stuck_tyfam_expansion =
                    (decsToTH [ -- type family F (x :: k) :: k
                                DOpenTypeFamilyD
                                  (DTypeFamilyHead fam_name
-                                                  [DKindedTV x () (DVarT k)]
+                                                  [DKindedTV x THAbs.BndrReq (DVarT k)]
                                                   (DKindSig (DVarT k))
                                                   Nothing)
                                -- type instance F (x :: ()) = x
@@ -331,7 +370,7 @@ test_getDataD_kind_sig =
                                   (DArrowT `DAppT` type_kind `DAppT` type_kind)
             (_, tvbs, _) <-
               withLocalDeclarations
-                [decToTH (DDataD Data [] data_name [DPlainTV a ()]
+                [decToTH (DDataD Data [] data_name [DPlainTV a THAbs.BndrReq]
                                  (Just data_kind_sig) [] [])]
                 (getDataD "th-desugar: Impossible" data_name)
             [| $(Syn.lift (length tvbs)) |])
@@ -410,8 +449,8 @@ test_t132 =
            --     m :: a
            --
            -- We define this by hand to avoid GHC#17608 on pre-9.0 GHCs.
-           decs = sweeten [ DClassD [] c [DPlainTV a ()] []
-                            [ DLetDec (DInfixD fixity m)
+           decs = sweeten [ DClassD [] c [DPlainTV a THAbs.BndrReq] []
+                            [ DLetDec (DInfixD fixity NoNamespaceSpecifier m)
                             , DLetDec (DSigD m (DVarT a))
                             ]
                           ]
@@ -574,7 +613,7 @@ test_t171 =
                    Data
                    []
                    t
-                   [DPlainTV x (), DPlainTV y ()]
+                   [DPlainTV x THAbs.BndrReq, DPlainTV y THAbs.BndrReq]
                    Nothing
                    [ DCon
                        [bTvb, aTvb, cTvb]
@@ -600,6 +639,99 @@ test_t171 =
          Just (DVarI _ actual_ty _) <- dsReify getT1
          expected_ty `eqTHSplice` actual_ty)
 
+-- Unit tests for tupleNameDegree_maybe. These also act as a regression test for
+-- #187.
+test_t187 :: [Bool]
+test_t187 =
+  map (\(s, expected) -> tupleNameDegree_maybe s == expected)
+    [ (''(),                Just 0)
+    , (''(,),               Just 2)
+    , (''(,,),              Just 3)
+    , (''Maybe,             Nothing)
+    , (tupleTypeName 0,     Just 0)
+    , (tupleTypeName 2,     Just 2)
+    , (tupleTypeName 3,     Just 3)
+#if __GLASGOW_HASKELL__ >= 900
+    , (''Solo,              Just 1)
+#if __GLASGOW_HASKELL__ >= 906
+    , ('MkSolo,             Just 1)
+#else
+    , ('Solo,               Just 1)
+#endif
+#endif
+#if __GLASGOW_HASKELL__ >= 908
+    , (''Unit,              Just 0)
+    , (''Tuple0,            Just 0)
+    , (''Tuple1,            Just 1)
+    , (''Tuple2,            Just 2)
+    , (''Tuple3,            Just 3)
+    , (''FakeTuples.Tuple0, Nothing)
+    , (''FakeTuples.Tuple1, Nothing)
+    , (''FakeTuples.Tuple2, Nothing)
+    , (''FakeTuples.Tuple3, Nothing)
+#endif
+    ]
+
+-- A regression test for #188, which ensures that it produces the correct answer
+-- for an unusual telescope like:
+--
+--   ... forall (a1 :: a2). forall (a2 :: a1). ...
+--
+-- Here, a2 is free in the kind of a1 (the first `forall`), but then the second
+-- `forall` binds another a2 that shadows what was already in scope. In this
+-- example, `toposortKindVarsOfTvbs [(a1 :: a2), (a2 :: a1)]` should return
+-- [a2].
+test_t188 :: Bool
+test_t188 =
+  let a1 = mkName "a1"
+      a2 = mkName "a2" in
+  toposortKindVarsOfTvbs [DKindedTV a1 () (DVarT a2), DKindedTV a2 () (DVarT a1)]
+    == [DPlainTV a2 ()]
+
+-- Unit tests for unboxedTupleNameDegree_maybe and unboxedSumNameDegree_maybe.
+-- These also act as a regression test for #213.
+test_t213 :: [Bool]
+test_t213 =
+  map (\(s, expected) -> unboxedTupleNameDegree_maybe s == expected)
+    [ (''(##),                 Just 0)
+    , (''(#,#),                Just 2)
+    , (''(#,,#),               Just 3)
+    , (''Maybe,                Nothing)
+#if __GLASGOW_HASKELL__ >= 802
+    , (unboxedTupleTypeName 0, Just 0)
+#endif
+    , (unboxedTupleTypeName 2, Just 2)
+    , (unboxedTupleTypeName 3, Just 3)
+#if __GLASGOW_HASKELL__ >= 910
+    , (''Unit#,                Just 0)
+    , (''Tuple0#,              Just 0)
+    , (''Solo#,                Just 1)
+    , (''Tuple1#,              Just 1)
+    , (''Tuple2#,              Just 2)
+    , (''Tuple3#,              Just 3)
+    , (''FakeTuples.Tuple0#,   Nothing)
+    , (''FakeTuples.Tuple1#,   Nothing)
+    , (''FakeTuples.Tuple2#,   Nothing)
+    , (''FakeTuples.Tuple3#,   Nothing)
+#endif
+    ]
+#if __GLASGOW_HASKELL__ >= 802
+  ++
+  map (\(s, expected) -> unboxedSumNameDegree_maybe s == expected)
+    [ (unboxedSumTypeName 2, Just 2)
+    , (unboxedSumTypeName 3, Just 3)
+    , (unboxedSumTypeName 4, Just 4)
+#if __GLASGOW_HASKELL__ >= 910
+    , (''Sum2#,              Just 2)
+    , (''Sum3#,              Just 3)
+    , (''Sum4#,              Just 4)
+    , (''FakeSums.Sum2#,     Nothing)
+    , (''FakeSums.Sum3#,     Nothing)
+    , (''FakeSums.Sum4#,     Nothing)
+#endif
+    ]
+#endif
+
 -- Unit tests for functions that compute free variables (e.g., fvDType)
 test_fvs :: [Bool]
 test_fvs =
@@ -740,9 +872,17 @@ test_matchTy =
   , matchTy NoIgnore (DConT ''Int) (DConT ''Bool) == Nothing
   , matchTy NoIgnore (DConT ''Int) (DConT ''Int) == Just M.empty
   , matchTy NoIgnore (DConT ''Int) (DVarT a) == Nothing
+    -- Test `DSigT` with both `IgnoreKinds` options
   , matchTy NoIgnore (DVarT a `DSigT` DConT ''Bool) (DConT ''Int) == Nothing
   , matchTy YesIgnore (DVarT a `DSigT` DConT ''Bool) (DConT ''Int)
     == Just (M.singleton a (DConT ''Int))
+    -- Test `DAppKindT` with both `IgnoreKinds` options
+  , matchTy NoIgnore (DConT ''Proxy `DAppKindT` DConT ''Bool `DAppT` DVarT a)
+                     (DConT ''Proxy `DAppT` DConT ''Int)
+    == Nothing
+  , matchTy YesIgnore (DConT ''Proxy `DAppKindT` DConT ''Bool `DAppT` DVarT a)
+                      (DConT ''Proxy `DAppT` DConT ''Int)
+    == Just (M.singleton a (DConT ''Int))
   ]
   where
     a = mkName "a"
@@ -823,6 +963,17 @@ main = hspec $ do
 
     it "locally reifies GADT record selector types with explicit foralls correctly" $ test_t171
 
+    it "doesn't reify a field selector with lookupValueNameWithLocals when NoFieldSelectors is set" $
+      t183 == Nothing
+
+    zipWithM (\b n -> it ("recognizes tuple names with tupleDegree_maybe correctly " ++ show n) b)
+      test_t187 [1..]
+
+    it "computes free kind variables correctly in a telescope that uses shadowing" $ test_t188
+
+    zipWithM (\b n -> it ("recognizes unboxed {tuple,sum} names with unboxed{Tuple,Sum}Degree_maybe correctly " ++ show n) b)
+      test_t213 [1..]
+
     -- Remove map pprints here after switch to th-orphans
     zipWithM (\t t' -> it ("can do Type->DType->Type of " ++ t) $ t == t')
              $(sequence round_trip_types >>= Syn.lift . map pprint)
@@ -830,8 +981,8 @@ main = hspec $ do
                mapM (\ t -> withLocalDeclarations [] (dsType t >>= expandType >>= return . typeToTH)) >>=
               Syn.lift . map pprint)
 
-    zipWith3M (\a b n -> it ("reifies local definition " ++ show n) $ a == b)
-      local_reifications normal_reifications [1..]
+    zipWith3M (\a b nm -> it ("reifies local definition " ++ nameBase nm) $ a == b)
+      local_reifications normal_reifications reifyDecsNames
 
     zipWithM (\b n -> it ("works on simplCase test " ++ show n) b) simplCase [1..]
 
