@@ -29,6 +29,10 @@ rae@cs.brynmawr.edu
 {-# LANGUAGE QuantifiedConstraints #-}
 #endif
 
+#if __GLASGOW_HASKELL__ < 806
+{-# LANGUAGE TypeInType #-}
+#endif
+
 #if __GLASGOW_HASKELL__ >= 807
 {-# LANGUAGE ImplicitParams #-}
 #endif
@@ -49,12 +53,22 @@ rae@cs.brynmawr.edu
 {-# LANGUAGE TypeData #-}
 #endif
 
+#if __GLASGOW_HASKELL__ >= 907
+{-# LANGUAGE TypeAbstractions #-}
+#endif
+
+#if __GLASGOW_HASKELL__ >= 909
+{-# LANGUAGE RequiredTypeArguments #-}
+#endif
+
 {-# OPTIONS_GHC -Wno-missing-signatures -Wno-type-defaults
                 -Wno-name-shadowing #-}
 
 module Splices where
 
 import qualified Data.List as L
+import qualified Data.List.NonEmpty as NE
+import Data.List.NonEmpty (NonEmpty(..))
 import Data.Char
 import qualified Data.Kind as Kind (Type)
 import GHC.Exts
@@ -94,16 +108,33 @@ unqualify :: Data a => a -> a
 unqualify = everywhere (mkT (mkName . nameBase))
 
 assumeStarT :: Data a => a -> a
-assumeStarT = everywhere (mkT assume_spec . mkT assume_unit)
+assumeStarT = everywhere (assume_spec_t . assume_vis_t . assume_unit_t)
   where
-    assume_spec :: TyVarBndrSpec -> TyVarBndrSpec
+    assume_spec_t :: Typeable a => a -> a
 #if __GLASGOW_HASKELL__ >= 900
+    assume_spec_t = mkT assume_spec
+
+    assume_spec :: TyVarBndrSpec -> TyVarBndrSpec
     assume_spec (PlainTV n spec)    = KindedTV n spec StarT
     assume_spec (KindedTV n spec k) = KindedTV n spec (assumeStarT k)
 #else
-    assume_spec = assume_unit
+    assume_spec_t = id
 #endif
 
+    assume_vis_t :: Typeable a => a -> a
+#if __GLASGOW_HASKELL__ >= 907
+    assume_vis_t = mkT assume_vis
+
+    assume_vis :: TyVarBndrVis -> TyVarBndrVis
+    assume_vis (PlainTV n vis)    = KindedTV n vis StarT
+    assume_vis (KindedTV n vis k) = KindedTV n vis (assumeStarT k)
+#else
+    assume_vis_t = id
+#endif
+
+    assume_unit_t :: Typeable a => a -> a
+    assume_unit_t = mkT assume_unit
+
     assume_unit :: TyVarBndrUnit -> TyVarBndrUnit
     assume_unit = elimTV (\n   -> kindedTV n StarT)
                          (\n k -> kindedTV n (assumeStarT k))
@@ -111,10 +142,11 @@ assumeStarT = everywhere (mkT assume_spec . mkT assume_unit)
 dropTrailing0s :: Data a => a -> a
 dropTrailing0s = everywhere (mkT (mkName . frob . nameBase))
   where
-    frob str
-      | head str == 'r' = str
-      | head str == 'R' = str
-      | otherwise       = L.dropWhileEnd isDigit str
+    frob str =
+      case str of
+        'r':_ -> str
+        'R':_ -> str
+        _     -> L.dropWhileEnd isDigit str
 
 -- Because th-desugar does not support linear types, we must pretend like
 -- MulArrowT does not exist for testing purposes.
@@ -158,11 +190,11 @@ test13_sig = [| show (read "[10, 11, 12]" :: [Int]) |]
 data Record = MkRecord1 { field1 :: Bool, field2 :: Int }
             | MkRecord2 { field2 :: Int, field3 :: Char }
 
-test14_record = [| let r1 = [MkRecord1 { field2 = 5, field1 = False }, MkRecord2 { field2 = 6, field3 = 'q' }]
-                       r2 = map (\r -> r { field2 = 18 }) r1
-                       r3 = (head r2) { field1 = True } in
-                   map (\case MkRecord1 { field2 = some_int, field1 = some_bool } -> show some_int ++ show some_bool
-                              MkRecord2 { field2 = some_int, field3 = some_char } -> show some_int ++ show some_char) (r3 : r2) |]
+test14_record = [| let r1 = MkRecord1 { field2 = 5, field1 = False } :| [MkRecord2 { field2 = 6, field3 = 'q' }]
+                       r2 = fmap (\r -> r { field2 = 18 }) r1
+                       r3 = (NE.head r2) { field1 = True } in
+                   fmap (\case MkRecord1 { field2 = some_int, field1 = some_bool } -> show some_int ++ show some_bool
+                               MkRecord2 { field2 = some_int, field3 = some_char } -> show some_int ++ show some_char) (NE.cons r3 r2) |]
 
 test15_litp = [| map (\case { 5 -> True ; _ -> False }) [5,6] |]
 test16_tupp = [| map (\(x,y,z) -> x + y + z) [(1,2,3),(4,5,6)] |]
@@ -201,8 +233,8 @@ test27_kisig = [| let f :: Proxy (a :: Bool) -> ()
 test28_tupt = [| let f :: (a,b) -> a
                      f (a,_) = a in
                  map f [(1,'a'),(2,'b')] |]
-test29_listt = [| let f :: [[a]] -> a
-                      f = head . head in
+test29_listt = [| let f :: [[Int]] -> [[Int]]
+                      f = map (map (+1)) in
                   map f [ [[1]], [[2]] ] |]
 test30_promoted = [| let f :: Proxy '() -> Proxy '[Int, Bool] -> ()
                          f _ _ = () in
@@ -355,6 +387,34 @@ test56_lambda_cases =
              _        _        -> "") (Just "Hello") (Just "World") |]
 #endif
 
+#if __GLASGOW_HASKELL__ >= 907
+test57_typed_th_bracket =
+  typedBracketE [| 'x' |]
+
+test58_typed_th_splice =
+  typedSpliceE (typedBracketE [| 'y' |])
+#endif
+
+#if __GLASGOW_HASKELL__ >= 909
+test59_embedded_types_keyword =
+  [| let idv :: forall a -> a -> a
+         idv (type a) (x :: a) = x :: a
+
+     in idv (type Bool) True |]
+
+test60_embedded_types_no_keyword =
+  [| let idv :: forall a -> a -> a
+         idv a (x :: a) = x :: a
+
+     in idv Bool True |]
+
+test61_invis_type_pat =
+  [| let f :: a -> a
+         f @a = id @a
+
+     in f @Bool True |]
+#endif
+
 type family TFExpand x
 type instance TFExpand Int = Bool
 type instance TFExpand (Maybe a) = [a]
@@ -504,6 +564,25 @@ dectest18 = [d| data Dec18 :: forall k -> k -> Kind.Type where
                   MkDec18 :: forall k (a :: k). Dec18 k a |]
 #endif
 
+#if __GLASGOW_HASKELL__ >= 907
+dectest19 = [d| type Dec19 :: forall k. k -> Kind.Type
+                data Dec19 @k (a :: k) = MkDec19 k (Proxy a) |]
+#endif
+
+#if __GLASGOW_HASKELL__ >= 909
+dectest20 = [d| infixr 3 data !@#
+                infixr 3 type !@#
+
+                (!@#) :: Bool -> Bool -> Bool
+                (!@#) = (&&)
+
+                type family (!@#) :: Bool -> Bool -> Bool |]
+
+dectest21 = [d| {-# SCC dec21 "dec21" #-}
+                dec21 :: a -> a
+                dec21 x = x |]
+#endif
+
 instance_test = [d| instance (Show a, Show b) => Show (a -> b) where
                        show _ = "function" |]
 
@@ -672,6 +751,13 @@ reifyDecs = [d|
   type data R38 a where
     R39 :: forall a. a -> R38 a
 #endif
+
+  -- A regression test for #184
+  data family x ^^^ y
+  data instance x ^^^ y = R40 x y
+
+  -- A regression test for #188
+  data R41 a (x :: Maybe a) = R42
   |]
 
 reifyDecsNames :: [Name]
@@ -690,6 +776,8 @@ reifyDecsNames = map mkName
 #if __GLASGOW_HASKELL__ >= 906
   , "R36", "R37", "R38", "R39"
 #endif
+  , "R40"
+  , "R41", "R42"
   ]
 
 simplCaseTests :: [Q Exp]
@@ -787,5 +875,14 @@ test_exprs = [ test1_sections
 #if __GLASGOW_HASKELL__ >= 903
              , test55_opaque_pragma
              , test56_lambda_cases
+#endif
+#if __GLASGOW_HASKELL__ >= 907
+             , test57_typed_th_bracket
+             , test58_typed_th_splice
+#endif
+#if __GLASGOW_HASKELL__ >= 909
+             , test59_embedded_types_keyword
+             , test60_embedded_types_no_keyword
+             , test61_invis_type_pat
 #endif
              ]
